#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include <time.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <SDL/SDL_ttf.h>


#define TS_TIMESTEP 100000
#define RAND_WAIT_MIN 500000
#define RAND_WAIT_MAX 1000000

typedef struct {
   SDL_sem * playEnded;
   int isEnabled;
   unsigned int timeCurr;
   unsigned int timeMax;
   char const * val;
} ToneCallbackData;

typedef struct {
	int listening;
	FILE *response_file;
} response_ctx_t;

typedef struct {
	int tau;
	int num_events;
	SDL_Surface *screen;
} stimulus_ctx_t;


static short const sineVals[] = {
   -9720, -9591, -9267, -8778, -8076, -7230, -6228, -5094, -3855, -2529, -1176,
   207, 1605, 2964, 4269, 5478, 6597, 7593, 8412, 9084, 9558, 9846, 9921, 9630,
   9342, 8778, 8127, 7299, 6291, 5202, 3999, 2700, 1350, -21, -1389, -2733,
   -4050, -5265, -6372, -7365, -8184, -8862, -9309, -9618
};

static char const * const sineValsMax = (char *)sineVals + sizeof(sineVals);

static void toneCallback(void * udata, Uint8 * stream, int len);
int playTone(int freq, int duration, ToneCallbackData *d);

int listen_for_responses(void *file_ptr);
void show_help(void);


int main(int argc, char *argv[])
{
	int i=0,c;
	int num_events = 100;
	int tau = 2;
	int wordDelay = 500000;
	int text_height, text_width;
	char *word = "X";
	FILE *stimulus_file = NULL;
	FILE *response_file = NULL;
	response_ctx_t *response_ctx = NULL;
	stimulus_ctx_t *stimulus_ctx = NULL;
	SDL_Surface *screen = NULL;
	SDL_Color clrWhite = { 255,255,255, 0 };
	SDL_Surface *sText = NULL;
	SDL_Thread *resp_thread = NULL;
	SDL_Thread *stim_thread = NULL;
	SDL_AudioSpec *waveSpec = NULL;
	ToneCallbackData *callbackData = NULL;

	// Command line parsing
	while( (c = getopt(argc, argv, "s:r:n:t:h")) != -1 ) {
		switch(c) {
			case 's':
				stimulus_file = fopen(optarg, "w");
				if( !stimulus_file ) {
					perror("stimulus file open");
					exit(1);
				}
				break;
			case 'r':
				response_file = fopen(optarg, "w");
				if( !response_file ) {
					perror("response file open");
					exit(1);
				}
				break;
			case 'n':
				num_events = atoi(optarg);
				if( num_events <= 0 ) {
					fprintf(stderr, "num_events (-n) must be a strictly positive integer.\n");
					exit(1);
				}
				break;
			case 't':
				tau = atoi(optarg);
				if( tau <= 0 ) {
					fprintf(stderr, "tau (-t) must be a strictly positive integer.\n");
					exit(1);
				}
				break;
			case '?':
			case 'h':
			default:
				show_help();
				exit(0);
		}
	}

	if( !stimulus_file ) {
		fprintf(stderr, "You must specify a stimulus output file.\n");
		exit(1);
	}
	if( !response_file ) {
		fprintf(stderr, "You must specify a response output file.\n");
		exit(1);
	}
	if( tau > num_events ) {
		fprintf(stderr, "Tau cannot be greater than num_events.\n");
		exit(1);
	}
	

	// Initialize Graphics
	if ( SDL_Init(SDL_INIT_AUDIO|SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(1);
	}
	atexit(SDL_Quit);

	// Initialize Audio
	waveSpec = malloc(sizeof(SDL_AudioSpec));
	callbackData = malloc(sizeof(ToneCallbackData));

	if( !waveSpec ) {
		fprintf(stderr, "Could not malloc audio spec.\n");
		exit(1);
	}
	if( !callbackData ) {
		fprintf(stderr, "Could not malloc callback data.\n");
		exit(1);
	}

	callbackData->isEnabled = 0;

	waveSpec->freq = 44100;  // Samples per sec (FM Radio quality)
	waveSpec->format = AUDIO_S16LSB;        // Signed 16-bit little-endian samples
	waveSpec->channels = 1;                 // Number of sound channels (stereo)
	waveSpec->silence = 0;
	waveSpec->samples = 4096;               // Size of the Audio buffer in samples
	waveSpec->padding = 0;
	waveSpec->size = 8192;
	waveSpec->callback = toneCallback;      // "Play a Tone" callback
	waveSpec->userdata = callbackData;                // waveCallback data

	if( SDL_OpenAudio(waveSpec, NULL) < 0 ) {
		fprintf(stderr, "Unable to open audio: %s\n", SDL_GetError()); 
		exit(1);
	}
	SDL_PauseAudio(0);

	screen = SDL_SetVideoMode(800, 600, 16, SDL_SWSURFACE);
	if ( screen == NULL ) {
		fprintf(stderr, "Unable to set 640x480 video: %s\n", SDL_GetError());
		exit(1);
	}

	// Initialize font library
	TTF_Init();
	atexit(TTF_Quit);

	// Set up sync mechanism
	event_sem = SDL_CreateSemaphore(0);

	// Create child processes
	//
	// Response thread
	response_ctx_t *response_ctx = malloc(sizeof(response_ctx_t));
	if( response_ctx == NULL ) {
		fprintf(stderr, "Unable to create response thread context data.\n");
		exit(1);
	}
	response_ctx->listening=1;
	response_ctx->response_file = response_file;
	resp_thread = SDL_CreateThread( response_thread, (void *)response_ctx );
	if ( resp_thread == NULL ) {
		fprintf(stderr, "Unable to create response thread: %s\n", SDL_GetError());
		exit(1);
	}

	// Stimulus thread
	stimulus_ctx = malloc(sizeof(response_ctx_t));
	if( stimulus_ctx == NULL ) {
		fprintf(stderr, "Unable to create stimulus thread context data.\n");
		exit(1);
	}
	stimulus_ctx->tau = tau;
	stimulus_ctx->num_events = num_events;
	stimulus_ctx->screen = screen;
	stimulus_ctx->stimulus_file = stimulus_file;
	stim_thread = SDL_CreateThread( stimulus_thread, (void *)stimulus_ctx );
	if ( stim_thread == NULL ) {
		fprintf(stderr, "Unable to create stimulus thread: %s\n", SDL_GetError());
		exit(1);
	}


	// Event Loop
	while ( SDL_WaitEvent(&event) ) {
		event_time = SDL_GetTicks();
		switch (event.type) {
			case SDL_KEYDOWN:
				if( event.key.keysym.sym == SDLK_q ) {
					fprintf(response_ctx->response_file, "User hit 'q' at %u\n", ticks);
					response_ctx->listening = 0;
					return 0;
				}

				// Save and signal
				response_ctx->last_event_time = event_time;
				SDL_SemPost(event_sem);
				fprintf(response_ctx->response_file, "%u\n", ticks);
				break;
			case SDL_QUIT:
				return 0;
				break;
		}
	}

	


	// Tell child to close
	// XXX this doesn't work yet (or anymore, whatever)
	SDL_Event event;
	event.type = SDL_QUIT;
	SDL_PushEvent( &event );
	SDL_WaitThread( resp_thread, NULL );


	// Clean up
	free(time_data);
	free(delay_data);
	TTF_CloseFont(font);
	fclose(stimulus_file);
	fclose(response_file);
	SDL_CloseAudio();
	free(callbackData);
	free(waveSpec);
	free(response_ctx);

	return 0;
}

int stimulus_thread(void *data)
{
	Uint32 max = 0;
	Uint32 min = ~0;
	Uint32 *delay_data = NULL;
	Uint32 *time_data = NULL;
	SDL_Rect textDest;
	TTF_Font *font = TTF_OpenFont( "./VeraMono.ttf", 36 );
	stimulus_ctx_t *stimulus_ctx = (stimulus_ctx *)data;

	// initialize random number generator
	srand(time(NULL));

	// Allocate buffer for times
	delay_data = malloc(num_events * sizeof(Uint32));
	if( !delay_data ) {
		fprintf(stderr, "Couldn't allocate memory for delay buffer.\n");
	}
	time_data = malloc(num_events * sizeof(Uint32));
	if( !time_data ) {
		fprintf(stderr, "Couldn't allocate memory for time buffer.\n");
		exit(1);
	}
	memset(time_data, 0, num_events * sizeof(Uint32));

	// Calculate delay times from a chaotic time series.
	delay_data[0] = rand();
	//	First do up to tau, since we can't go all the way back yet.
	for( i=1; i<tau; i++ ) {
		delay_data[i] = delay_data[i-1] + TS_TIMESTEP * (-1*delay_data[i-1] - 20*sin(delay_data[0]));
		if( delay_data[i] > max ) {
			max = delay_data[i];
		} else if( delay_data[i] < min ) {
			min = delay_data[i];
		}
	}
	//	Then do the rest
	for( i=tau; i<num_events; i++ ) {
		delay_data[i] = delay_data[i-1] + TS_TIMESTEP * (-1*delay_data[i-1] - 20*sin(delay_data[i-1-tau]));
		if( delay_data[i] > max ) {
			max = delay_data[i];
		} else if( delay_data[i] < min ) {
			min = delay_data[i];
		}
	}
	//	Remap the data to fit between 0 and 1 second
	for( i=0; i<num_events; i++ ) {
		delay_data[i] -= min;
		delay_data[i] /= (max - min)/1000000;
	}

	// This will center the text on the screen
	TTF_SizeText(font, word, &text_width, &text_height);
	tx1 = (screen->w)/2 - text_width/2;
	ty1 = (screen->h)/2 - text_height/2;
	tx2 = (screen->w)/2 + text_width/2;
	ty2 = (screen->h)/2 + text_height/2;
	textDest.x = tx1; textDest.y = txy1;
	textDest.w = tx2-tx1; textDest.h = ty2-ty1;


	// Provide the stimulus
	// only essential things go in this loop
	for( i=0; i<num_events && response_ctx->listening; i++) {

		usleep( RAND_WAIT_MIN + delay_data[i] % RAND_WAIT_MAX );

		playTone(262, 500, callbackData);
		// Display word...
		sText = TTF_RenderText_Solid(font, word, clrWhite);
		SDL_BlitSurface(sText, NULL, screen, &textDest );
		SDL_FreeSurface(sText);
		SDL_UpdateRect(screen, tx1,ty1, tx2,ty2 );
		time_data[i] = SDL_GetTicks();

		// ...for this long
		usleep(wordDelay);

		// Blank
		SDL_FillRect(screen, NULL, 0x000000);
		SDL_UpdateRect(screen, tx1,ty1, tx2,ty2 );
	}

	// Write data out to the stimulus file
	for( i=0; i<num_events; i++ ) {
		fprintf(stimulus_file, "%u\t%u\n", time_data[i], delay_data[i]);
	}

	return 0;
}


// takes a void * to conform to SDL_CreateThread
int response_thread(void *data)
{
	Uint32 ticks;
	SDL_Event event;
	response_ctx_t *response_ctx = (response_ctx_t *)data;

	while ( SDL_WaitEvent(&event) ) {
		switch (event.type) {
			case SDL_KEYDOWN:
				ticks = SDL_GetTicks();
				if( event.key.keysym.sym == SDLK_q ) {
					fprintf(response_ctx->response_file, "User hit 'q' at %u\n", ticks);
					response_ctx->listening = 0;
					return 0;
				}
				fprintf(response_ctx->response_file, "%u\n", ticks);
				break;
			case SDL_QUIT:
				return 0;
				break;
		}
	}

	// we should never get here
	return 1;
}

static void toneCallback(void * udata, Uint8 * stream, int len)
{
   ToneCallbackData *d = (ToneCallbackData *)udata;
   if (d->isEnabled) {
      if (d->timeCurr < d->timeMax) {
         if (d->timeCurr+len > d->timeMax) {
            len = d->timeMax - d->timeCurr;
         }
         d->timeCurr += len;
         len /= 2;
         while (len--) {
            if (d->val >= sineValsMax) {
               d->val = (char const *)sineVals;
            }
            *stream++ = *(d->val++);
            *stream++ = *(d->val++);
         }
         d->val = d->val;
      } else {
         d->isEnabled = 0;   /* disable self */
         SDL_SemPost(d->playEnded);
      }
   }
}


int playTone(int freq, int duration, ToneCallbackData *d)
{
	double ratio = freq / 1000.0;
	int rc = 0;


	// disable our toneCallback
	d->isEnabled = 0;

	d->playEnded = SDL_CreateSemaphore(0);
	d->timeCurr = 0;
	d->timeMax = (int)(82 * duration * ratio);
	d->val = (char const *)sineVals;
	d->isEnabled = 1;

	// wait here until the callback finishes
	SDL_SemWait(d->playEnded);
	SDL_DestroySemaphore(d->playEnded);

	return rc;
}

void show_help()
{
	printf("metronome -s <file> -r <file> -n <num> -t <tau> | [-h]\n\n");
	printf("  -s file  \tWrite metronome (stimulus) times to <file> (no default)\n");
	printf("  -r file  \tWrite keypress (response) times to <file> (no default)\n");
	printf("  -n num   \tNumber of stimulus events to produce (default: 100)\n");
	printf("  -t tau   \tAmount of delay used to produce delay times (default: 2)\n");
	printf("  -h	   \tHelp\n\n");
}

