/*
	Anynome - The Generalized Metronome

    Copyright (C) 2007  Nigel D. Stepp

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Nigel Stepp <stepp@atistar.net>

    $Id: anynome.c 146 2007-09-08 02:55:15Z stepp $
*/


#include <stdio.h>
#include <stdlib.h>

#ifndef WIN32
 #include <unistd.h>
#endif

#include <limits.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>
#include <time.h>

#ifndef WIN32
 #include <SDL/SDL.h>
 #include <SDL/SDL_thread.h>
 #include <SDL/SDL_ttf.h>
#else
 #include <Windows.h>
 #include "SDL.h"
 #include "SDL_thread.h"
 #include "SDL_ttf.h"
 #include "getopt.h"
#endif

#include "anynome.h"
#include "plugin.h"
#include "sound.h"

int main(int argc, char *argv[])
{
	int c;
	int video_flags;
	Uint32 event_time;
	FILE *stimulus_file = NULL;
	FILE *response_file = NULL;
	config_t *config = NULL;
	plugin_t *stimulus_plugin = NULL;
	stimulus_ctx_t *stimulus_ctx = NULL;
	sound_t *sound = NULL;
	SDL_Surface *screen = NULL;
	SDL_Event event;
	SDL_Thread *stim_thread = NULL;

	// Command line parsing
	
	if( argc == 1 ) {
		show_help();
		exit(1);
	}

	// set defaults
	config = malloc(sizeof(config_t));
	if( !config ) {
		fprintf(stderr, "Could not malloc config.\n");
		exit(1);
	}
	config->num_events = 100;
	config->use_audio = 0;
	config->fullscreen = 0;

	while( (c = getopt(argc, argv, "aftvhs:r:n:d:p:")) != -1 ) {
		switch(c) {
			case 's':
				stimulus_file = fopen(optarg, "w");
				if( !stimulus_file ) {
					perror("stimulus file open");
					exit(1);
				}
				break;
			case 'r':
				response_file = fopen(optarg, "w");
				if( !response_file ) {
					perror("response file open");
					exit(1);
				}
				break;
			case 'n':
				config->num_events = atoi(optarg);
				if( config->num_events <= 0 ) {
					fprintf(stderr, "num_events (-n) must be a strictly positive integer.\n");
					exit(1);
				}
				break;
			case 'd':
				strncpy(config->plugin_dir, optarg, sizeof(config->plugin_dir)-1);
				break;
			case 'p':
				strncpy(config->plugin, optarg, sizeof(config->plugin)-1);
				break;
			case 'a':
				config->use_audio = 1;
				break;
			case 'f':
				config->fullscreen = 1;
				break;
			case 't':
				timer_test();
				exit(0);
				break;
			case 'v':
				show_version();
				exit(0);
				break;
			case '?':
			case 'h':
			default:
				show_help();
				exit(0);
		}
	}


	// Handle plugins
	stimulus_plugin = plugin_load( config->plugin_dir, config->plugin );
	if( !stimulus_plugin ) {
		fprintf(stderr, "Error loading plugin.\n");
		exit(1);
	}

	// Send args to the stimulus plugin
	if( stimulus_plugin ) {
		plugin_set_args( stimulus_plugin, argc, argv, optind );
	}

	// Make sure of required parameters
	if( !stimulus_file ) {
		fprintf(stderr, "You must specify a stimulus output file.\n");
		exit(1);
	}
	if( !response_file ) {
		fprintf(stderr, "You must specify a response output file.\n");
		exit(1);
	}
	
	// Initialize Graphics
	if ( SDL_Init(SDL_INIT_AUDIO|SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(1);
	}
	atexit(SDL_Quit);

	// Initialize Audio
	if( config->use_audio ) {
		sound = init_sound();
		if( !sound ) {
			config->use_audio = 0;
		}
	}

	// Initialize font library
	TTF_Init();
	atexit(TTF_Quit);

	// Create a screen to draw on
	video_flags = SDL_SWSURFACE;
	if( config->fullscreen ) {
		video_flags |= SDL_FULLSCREEN;
	}
	screen = SDL_SetVideoMode(800, 600, 16, video_flags);
	if ( screen == NULL ) {
		fprintf(stderr, "Unable to set 800x600 video: %s\n", SDL_GetError());
		exit(1);
	}

	// Create child process
	//
	// Stimulus thread
	stimulus_ctx = malloc(sizeof(stimulus_ctx_t));
	if( stimulus_ctx == NULL ) {
		fprintf(stderr, "Unable to create stimulus thread context data.\n");
		exit(1);
	}
	stimulus_ctx->num_events = config->num_events;
	stimulus_ctx->use_audio = config->use_audio;
	stimulus_ctx->screen = screen;
	stimulus_ctx->sound = sound;
	stimulus_ctx->stimulus_file = stimulus_file;
	stimulus_ctx->done = 0;
	stimulus_ctx->get_delay_series = plugin_get_fcn(stimulus_plugin, "get_delay_series");
	stim_thread = SDL_CreateThread( stimulus_thread, (void *)stimulus_ctx );
	if ( stim_thread == NULL ) {
		fprintf(stderr, "Unable to create stimulus thread: %s\n", SDL_GetError());
		exit(1);
	}


	// Event Loop
	while ( !stimulus_ctx->done ) {
		if( !SDL_PollEvent(&event) ) {
			usleep(1000);
			continue;
		}
		event_time = SDL_GetTicks();
		switch (event.type) {
			case SDL_KEYDOWN:
				if( event.key.keysym.sym == SDLK_q ) {
					fprintf(response_file, "User hit 'q' at %u\n", event_time);
					stimulus_ctx->done = 1;
				} else {
					fprintf(response_file, "%u\n", event_time);
				}

				break;
			case SDL_QUIT:
				stimulus_ctx->done = 1;
				break;
		}
	}

	
	// Wait for the child to close
	SDL_WaitThread( stim_thread, NULL );


	// Clean up
	plugin_unload(stimulus_plugin);
	fclose(stimulus_file);
	fclose(response_file);
	close_sound(sound);
	free(stimulus_ctx);
	free(config);

	return 0;
}

int stimulus_thread(void *data)
{
	int i;
	Uint32 stimDelay = 500000;
	Uint32 delay;
	Uint32 *delay_data = NULL;
	Uint32 *time_data = NULL;
	SDL_Rect stimDest;
	TTF_Font *font = TTF_OpenFont( "./VeraMono.ttf", 36 );
	stimulus_ctx_t *stimulus_ctx = (stimulus_ctx_t *)data;

	// Allocate buffer for times
	delay_data = malloc(stimulus_ctx->num_events * sizeof(Uint32));
	if( !delay_data ) {
		fprintf(stderr, "Couldn't allocate memory for delay buffer.\n");
	}
	time_data = malloc(stimulus_ctx->num_events * sizeof(Uint32));
	if( !time_data ) {
		fprintf(stderr, "Couldn't allocate memory for time buffer.\n");
		exit(1);
	}
	memset(time_data, 0, stimulus_ctx->num_events * sizeof(Uint32));

	
	stimulus_ctx->get_delay_series(delay_data, stimulus_ctx->num_events);
	
	// This will center the square on the screen
	stimDest.x = (stimulus_ctx->screen->w)/2 - 50;
	stimDest.y = (stimulus_ctx->screen->h)/2 - 50;
	stimDest.w = 100;
	stimDest.h = 100;


	// Provide the stimulus
	// only essential things go in this loop
	for( i=0; i<stimulus_ctx->num_events && !stimulus_ctx->done; i++) {

		// Adjust delays to include stimulus display time,
		// also clamp to stimDelay so stimuli don't overlap.
		if( delay_data[i] > stimDelay ) {
			delay = delay_data[i] - stimDelay;
		} else {
			delay = 0;
		}

		usleep( delay );

		if( stimulus_ctx->use_audio ) {
			playTone(262, 500, stimulus_ctx->sound);
		}

		// Display stimulus...
		SDL_FillRect(stimulus_ctx->screen, &stimDest, 0xFFFFFF);
		SDL_UpdateRect(stimulus_ctx->screen, stimDest.x, stimDest.y, stimDest.w, stimDest.h);

		time_data[i] = SDL_GetTicks();

		// ...for this long
		usleep(stimDelay);

		// Blank
		SDL_FillRect(stimulus_ctx->screen, &stimDest, 0x000000);
		SDL_UpdateRect(stimulus_ctx->screen, stimDest.x, stimDest.y, stimDest.w, stimDest.h);
	}

	// Write data out to the stimulus file
	for( i=0; i<stimulus_ctx->num_events; i++ ) {
		fprintf(stimulus_ctx->stimulus_file, "%u\t%u\n", time_data[i], delay_data[i]);
	}

	free(time_data);
	free(delay_data);
	TTF_CloseFont(font);

	// Let everyone know we are done
	stimulus_ctx->done = 1;

	return 0;
}

void timer_test()
{
	int i;
	Uint32 off;
	Uint32 ticks[100];
	Uint32 *tick;
	SDL_Thread *thread;
	SDL_Event event;
	SDL_Surface *screen;

	if ( SDL_Init(0) < 0 ) {
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(1);
	}
	atexit(SDL_Quit);

	// We need this for events
	screen = SDL_SetVideoMode(80, 60, 16, SDL_SWSURFACE);
	if ( screen == NULL ) {
		fprintf(stderr, "Unable to set 800x600 video: %s\n", SDL_GetError());
		exit(1);
	}
	/* 
	 * Delay Timing
	 */
	printf("Testing 100ms delay resolution....\n");
	
	tick = ticks;
	for(i=0; i<100; i++) {
		usleep(100000);
		*(tick++) = SDL_GetTicks();
	}
	off = 0;
	for(i=1; i<100; i++) {
		off += ticks[i]-ticks[i-1] - 100;
	}
	printf("  Mean deviation after 100 delays: %fms\n", ((double)off)/99.0);

	printf("Testing 10ms delay resolution....\n");
	
	tick = ticks;
	for(i=0; i<100; i++) {
		usleep(10000);
		*(tick++) = SDL_GetTicks();
	}
	off = 0;
	for(i=1; i<100; i++) {
		off += ticks[i]-ticks[i-1] - 10;
	}
	printf("  Mean deviation after 100 delays: %fms\n", ((double)off)/99.0);

	printf("Testing 1ms delay resolution....\n");

	tick = ticks;
	for(i=0; i<100; i++) {
		usleep(1000);
		*(tick++) = SDL_GetTicks();
	}
	off = 0;
	for(i=1; i<100; i++) {
		off += ticks[i]-ticks[i-1] - 1;
	}
	printf("  Mean deviation after 100 delays: %fms\n", ((double)off)/99.0);

	/*
	 * Event Timing
	 */
	printf("\nTesting event resolution....\n\n");
	thread = SDL_CreateThread(event_thread, NULL);
	tick = ticks;
	for(i=0; i<100; ) {
		if( SDL_PollEvent(&event) && event.type == SDL_USEREVENT ) {
			*(tick++) = SDL_GetTicks()-event.user.code;
			i++;
		}
		usleep(1000);
	}
	if( !i ) {
		printf("i==0! (%s)\n", SDL_GetError());
		exit(1);
	}

	off=0;
	for(i=0; i<100; i++) {
		off += ticks[i];
	}
	printf("\n  Mean delay between 100 events: %fms\n\n", ((double)off)/100.0);
	SDL_WaitThread(thread, NULL);
}

int event_thread(void *data)
{
	int i;
	SDL_Event event;

	event.type = SDL_USEREVENT;

	for(i=0; i<100; i++) {
		usleep(1000);
		event.user.code=SDL_GetTicks();
		SDL_PushEvent(&event);
	}

	return 0;
}

void show_help()
{
	printf("anynome -aftvh -s <file> -r <file> -n <num> -p <plugin> [-- plugin args]\n\n");
	printf("  -a       \tBeep at each stimulus event (default: off)\n");
	printf("  -f       \tUse a fullscreen video mode (default: off)\n");
	printf("  -t       \tTiming tester\n");
	printf("  -v       \tShow version numbers\n");
	printf("  -h       \tHelp\n");
	printf("  -s file  \tWrite metronome (stimulus) times to <file> (default: stim.txt)\n");
	printf("  -r file  \tWrite keypress (response) times to <file> (default: resp.txt)\n");
	printf("  -n num   \tNumber of stimulus events to produce (default: 100)\n");
	printf("  -p plugin\tName of the stimulus plugin to use (default: constant)\n\n");
	printf("  Plugin Specific parameters may be listed after \"--\".\n");
	printf("  e.g. anynome -s s.txt -r r.txt -p ikeda -- -t 10\n");
	printf("       which uses the \"ikeda\" plugin passing \"-t 10\"\n");
	printf("       as a parameter (tau = 10 samples).\n\n");
	printf("  Use \"-- -h\" to list plugin parameters.\n");
	printf("\n");
}

void show_version()
{
	printf("Anynome version: %s\n", VERSION);
	printf("Plugin API version: %s\n", APIVERSION);
}


