/*
	Anynome - The Generalized Metronome

    Copyright (C) 2007  Nigel D. Stepp

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Nigel Stepp <stepp@atistar.net>

    $Id: plugin.c 146 2007-09-08 02:55:15Z stepp $
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#ifdef WIN32
 #include <windows.h>
 #define dlopen(a,b) (LoadLibrary(a))
 #define dlsym(a,b) (GetProcAddress(a,b))
 #define dlclose(a) (FreeLibrary(a))
 #define dlerror() (NULL)
 #define PATH_MAX MAX_PATH
#else
 #include <dlfcn.h>
#endif

#include "plugin.h"

plugin_t *plugin_load(char *plugin_dir, char *plugin_name)
{
	char plugin_filename[PATH_MAX];
	char api_version[8];
	void *plugin_handle = NULL;
	plugin_t *new_plugin = NULL;
	const char *(*get_api_version)(void) = NULL;

	memset(plugin_filename, 0, PATH_MAX);

#ifdef WIN32
	_snprintf(plugin_filename, PATH_MAX-1, "plugin_%s.dll", plugin_name);
#else
	snprintf(plugin_filename, PATH_MAX-1, "plugin_%s.so", plugin_name);
#endif

	plugin_handle = dlopen(plugin_filename, RTLD_NOW);
	if( !plugin_handle ) {
		fprintf(stderr, "Failed to load plugin library: %s\n", dlerror());
		exit(1);
	}

	new_plugin = malloc(sizeof(plugin_t));
	if( !new_plugin ) {
		fprintf(stderr, "Could not malloc new plugin.\n");
		exit(1);
	}

	new_plugin->handle = plugin_handle;
	new_plugin->name = strdup(plugin_name);

	get_api_version = plugin_get_fcn(new_plugin, "get_api_version");

	strncpy(api_version, get_api_version(), 7);
	api_version[7] = '\0';

	if( strcmp( APIVERSION, api_version ) ) {
		fprintf(stderr, "API versions do not match.  I am %s, plugin is %s.\n", APIVERSION, api_version);
		exit(1);
	}

	return new_plugin;
}

void plugin_unload(plugin_t *plugin)
{
	free(plugin->name);
	dlclose(plugin->handle);
	free(plugin);
}

void plugin_set_args(plugin_t *plugin, int argc, char **argv, int opt_ind)
{
	void (*set_args)(int, char **, int);

	set_args = plugin_get_fcn(plugin, "set_args");

	set_args(argc, argv, opt_ind);
}

void *plugin_get_fcn(plugin_t *plugin, const char *fcn)
{
	char *error;
	void *func;

	// clear errors. save so the Win32 def will work out better
	error = dlerror();
	func = dlsym(plugin->handle, fcn);
	if( (error = dlerror()) != NULL || !func ) {

#ifdef WIN32
		FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_FROM_SYSTEM |
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			GetLastError(),
			MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
			(LPTSTR) &error,
			0, NULL );
#endif
		fprintf(stderr, "Error getting plugin function '%s': %s\n", fcn, error);
		exit(1);
	}

	return func;
}

