/*
	Anynome - The Generalized Metronome

    Copyright (C) 2007  Nigel D. Stepp

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Nigel Stepp <stepp@atistar.net>

    $Id: plugin_ikeda.c 132 2007-08-30 01:38:47Z stepp $
*/


#include "plugin_common.h"

#define PLUGIN_LABEL "ikeda"

#define TS_TIMESTEP 100000
#define RAND_WAIT_MIN 500000
#define RAND_WAIT_MAX 1000000

// Define plugin parameters here
struct {
	int alpha;
	int beta;
	int tau;
} plugin_config;


void set_args(int argc, char *argv[], int opt_ind)
{
	int c;
	
	// set defaults
	plugin_config.alpha = -1;
	plugin_config.beta = 20;
	plugin_config.tau = 2;

	// Some getopt implementations don't preserve optind
	if( optind < opt_ind ) {
		optind = opt_ind;
	}
	while( (c = getopt(argc, argv, "a:b:t:h")) != -1 ) {
		switch(c) {
			case 'a':
				plugin_config.alpha = atoi(optarg);
				break;
			case 'b':
				plugin_config.beta = atoi(optarg);
				break;
			case 't':
				plugin_config.tau = atoi(optarg);
				if( plugin_config.tau <= 0 ) {
					plugin_error(PLUGIN_LABEL, "tau (-t) must be a strictly positive integer.");
					exit(1);
				}
				break;
			case '?':
			case 'h':
			default:
				show_help();
				exit(0);
		}
	}
}


void get_delay_series(uint32_t *delay_data, int num_events)
{
	int i;
	uint32_t max = 0;
	uint32_t min = ~0;

	if( !delay_data || num_events <= 0 ) {
		return;
	}
	if( plugin_config.tau >= num_events ) {
		plugin_error(PLUGIN_LABEL, "tau must be less than the number of stimulus events.");
		exit(1);
	}

	// initialize random number generator
	srand(time(NULL));

	delay_data[0] = rand();

	//	First do up to tau, since we can't go all the way back yet.
	for( i=1; i<plugin_config.tau; i++ ) {
		delay_data[i] = delay_data[i-1] + TS_TIMESTEP * ( plugin_config.alpha*delay_data[i-1] - plugin_config.beta*sin(delay_data[0]) );
		if( delay_data[i] > max ) {
			max = delay_data[i];
		} else if( delay_data[i] < min ) {
			min = delay_data[i];
		}
	}

	//	Then do the rest
	for( i=plugin_config.tau; i<num_events; i++ ) {
		delay_data[i] = delay_data[i-1] + TS_TIMESTEP * ( plugin_config.alpha*delay_data[i-1] - plugin_config.beta*sin(delay_data[i-1-plugin_config.tau]) );
		if( delay_data[i] > max ) {
			max = delay_data[i];
		} else if( delay_data[i] < min ) {
			min = delay_data[i];
		}
	}

	//	Remap the data to fit between 1 and 2 seconds
	for( i=0; i<num_events; i++ ) {
		delay_data[i] -= min;
		delay_data[i] /= (max - min)/1000000;
		delay_data[i] += 1000000;
	}
}

void show_help(void)
{
	printf("Plugin Parameters (%s):\n", PLUGIN_LABEL);
	printf("  -a alpha \tIkeda map alpha parameter (default: -1)\n");
	printf("  -b beta  \tIkeda map beta parameter (default: 20)\n");
	printf("  -t delay \tDelay to use for delayed feedback (default: 2)\n");
	printf("  -h       \tPlugin help\n");
	printf("\n");
}

