/*
	Anynome - The Generalized Metronome

    Copyright (C) 2007  Nigel D. Stepp

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    Nigel Stepp <stepp@atistar.net>

    $Id: sound.c 132 2007-08-30 01:38:47Z stepp $
*/


#include <stdio.h>
#include <stdlib.h>

#ifndef WIN32
 #include <unistd.h>
#endif

#include <malloc.h>
#include <math.h>
#include <time.h>

#ifndef WIN32
 #include <SDL/SDL.h>
#else
 #include <Windows.h>
 #include "SDL.h"
#endif

#include "sound.h"

static void toneCallback(void * udata, Uint8 * stream, int len);

static short const sineVals[] = {
   -9720, -9591, -9267, -8778, -8076, -7230, -6228, -5094, -3855, -2529, -1176,
   207, 1605, 2964, 4269, 5478, 6597, 7593, 8412, 9084, 9558, 9846, 9921, 9630,
   9342, 8778, 8127, 7299, 6291, 5202, 3999, 2700, 1350, -21, -1389, -2733,
   -4050, -5265, -6372, -7365, -8184, -8862, -9309, -9618
};

static char const * const sineValsMax = (char *)sineVals + sizeof(sineVals);


sound_t *init_sound(void)
{
	sound_t *sound = NULL;
	SDL_AudioSpec *waveSpec = NULL;
	ToneCallbackData *callbackData = NULL;

	// Initialize Audio
	sound = malloc(sizeof(sound_t));
	waveSpec = malloc(sizeof(SDL_AudioSpec));
	callbackData = malloc(sizeof(ToneCallbackData));

	if( !sound ) {
		fprintf(stderr, "Could not malloc sound context.\n");
		exit(1);
	}
	if( !waveSpec ) {
		fprintf(stderr, "Could not malloc audio spec.\n");
		exit(1);
	}
	if( !callbackData ) {
		fprintf(stderr, "Could not malloc callback data.\n");
		exit(1);
	}

	callbackData->isEnabled = 0;
	callbackData->playEnded = SDL_CreateSemaphore(0);

	waveSpec->freq = 44100;  // Samples per sec (FM Radio quality)
	waveSpec->format = AUDIO_S16LSB;        // Signed 16-bit little-endian samples
	waveSpec->channels = 1;                 // Number of sound channels (stereo)
	waveSpec->silence = 0;
	waveSpec->samples = 4096;               // Size of the Audio buffer in samples
	waveSpec->padding = 0;
	waveSpec->size = 8192;
	waveSpec->callback = toneCallback;      // "Play a Tone" callback
	waveSpec->userdata = callbackData;                // waveCallback data

	if( SDL_OpenAudio(waveSpec, NULL) < 0 ) {
		fprintf(stderr, "Unable to open audio: %s\n", SDL_GetError());
		SDL_DestroySemaphore(callbackData->playEnded);
		free(callbackData);
		free(waveSpec);
		free(sound);
		return NULL;
	}
	SDL_PauseAudio(0);

	sound->waveSpec = waveSpec;
	sound->callbackData = callbackData;

	return sound;
}

void close_sound(sound_t *sound)
{
	if( !sound ) {
		return;
	}

	SDL_DestroySemaphore(sound->callbackData->playEnded);
	SDL_CloseAudio();
	free(sound->callbackData);
	free(sound->waveSpec);
	free(sound);
}

static void toneCallback(void * udata, Uint8 * stream, int len)
{
   ToneCallbackData *d = (ToneCallbackData *)udata;
   if (d->isEnabled) {
      if (d->timeCurr < d->timeMax) {
         if (d->timeCurr+len > d->timeMax) {
            len = d->timeMax - d->timeCurr;
         }
         d->timeCurr += len;
		 // len /= 2;
         len >>= 1;
         while (len--) {
            if (d->val >= sineValsMax) {
               d->val = (char const *)sineVals;
            }
            *stream++ = *(d->val++);
            *stream++ = *(d->val++);
         }
      } else {
         d->isEnabled = 0;   /* disable self */
         SDL_SemPost(d->playEnded);
      }
   }
}


int playTone(int freq, int duration, sound_t *sound)
{
	double ratio = freq / 1000.0;
	int rc = 0;

	if( !sound ) {
		return 1;
	}

	// disable our toneCallback
	sound->callbackData->isEnabled = 0;

	sound->callbackData->timeCurr = 0;
	sound->callbackData->timeMax = (int)(82 * duration * ratio);
	sound->callbackData->val = (char const *)sineVals;
	sound->callbackData->isEnabled = 1;

	// wait here until the callback finishes
	SDL_SemWait(sound->callbackData->playEnded);

	return rc;
}

